/***********************************************************************************************************************
 *
 * blueBill Mobile - Android - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it/mobile
 * SCM: https://java.net/hg/bluebill-mobile~android-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.mobile.news.spi;

import java.util.concurrent.ExecutorService;
import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.inject.Provider;
import java.util.Collection;
import java.util.Collections;
import java.io.IOException;
import java.net.URL;
import it.tidalwave.util.Task;
import it.tidalwave.util.thread.annotation.ThreadConfined;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ServiceProvider;
import it.tidalwave.bluebill.mobile.news.NewsPreferences;
import it.tidalwave.bluebill.mobile.news.NewsService;
import it.tidalwave.bluebill.mobile.news.impl.DocumentReadable;
import it.tidalwave.bluebill.mobile.news.impl.RssFeedReadable;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.netbeans.capabilitiesprovider.CapabilitiesProvider;
import it.tidalwave.netbeans.capabilitiesprovider.CapabilitiesProviderSupport;
import it.tidalwave.netbeans.capabilitiesprovider.ThreadLookupBinder;
import it.tidalwave.semantic.document.Document;
import it.tidalwave.semantic.rss.RssFeed;
import it.tidalwave.bluebill.mobile.news.impl.Cache;
import it.tidalwave.bluebill.mobile.news.impl.ReadMessagesLog;
import it.tidalwave.util.NotFoundException;
import lombok.extern.slf4j.Slf4j;
import static it.tidalwave.util.thread.ThreadType.*;
import static it.tidalwave.bluebill.mobile.news.impl.Cache.Status.*;

/***********************************************************************************************************************
 *
 * 
 * @stereotype Service
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@ServiceProvider(service=NewsService.class) @Slf4j
public class DefaultNewsService implements NewsService
  {
    private static final Class<DefaultNewsService> _ = DefaultNewsService.class;

    private static final String NEWS_FEED_URL = "http://bluebill.tidalwave.it/mobile/blog.rss";

    @CheckForNull
    protected Thread startThread;

    @Nonnull
    private final Provider<NewsPreferences> preferences = Locator.createProviderFor(NewsPreferences.class);

    /* package */ final Cache cache; 
    
    /* package */ final ReadMessagesLog readMessagesLog;

    @Nonnull
    private final ExecutorService executorService = Locator.find(ExecutorService.class);

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private final CapabilitiesProvider<?> c1 = new CapabilitiesProviderSupport<RssFeed>()
      {
        @Override @Nonnull
        public Collection<? extends Object> createCapabilities (final @Nonnull RssFeed rssFeed) 
          {
            return Collections.singletonList(new RssFeedReadable(rssFeed, readMessagesLog));
          }
      }; 
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private final CapabilitiesProvider<?> c2 = new CapabilitiesProviderSupport<Document>()
      {
        @Override @Nonnull
        public Collection<? extends Object> createCapabilities (final @Nonnull Document message) 
          {
            return Collections.singletonList(new DocumentReadable(message, readMessagesLog));
          }
      }; 
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @ThreadConfined(type=ANY)
    public DefaultNewsService()
      throws IOException
      {
        this.cache = new Cache(createUrl(NEWS_FEED_URL));
        readMessagesLog = new ReadMessagesLog();
      }

    /*******************************************************************************************************************
     *
     * For tests only.
     *
     ******************************************************************************************************************/
    @ThreadConfined(type=ANY)
    /* package */ DefaultNewsService (final @Nonnull Cache cache, final @Nonnull ReadMessagesLog readMessagesLog)
      {
        this.cache = cache;
        this.readMessagesLog = readMessagesLog;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override
    public synchronized void eventuallyCheckForUpdates()
      {
        log.info("eventuallyCheckForUpdates()");
        
        if (startThread == null)
          {
            if (!preferences.get().isNewsDownloadAllowed())
              {
                log.info(">>>> not checking, internet connections are disabled");
              }
            else
              {
                startThread = new Thread()
                  {
                    @Override
                    public void run()
                      {
                        try
                          {
                            updateCache();
                          }
                        catch (Exception e)
                          {
                            log.warn("While running news feed check", e);
                          }

                        startThread = null;
                      }
                  };

                startThread.start();
              }
          }
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    //@bluebook-begin getNewsFeed
    @Override
    public void getNewsFeed (final @Nonnull AvailabilityNotifier availabilityNotifier)
      {
        log.info("getNewsFeed({})", availabilityNotifier);

        try
          {
            availabilityNotifier.notifyFeedAvailable(cache.getRssFeed());
          }
        catch (NotFoundException e)
          {
            ensureCacheIsInitialized(availabilityNotifier);
            
            if (!cache.getStatus().isDownloadNeeded())
              {
                readNewsFeedAndNotifyAvailability(availabilityNotifier);                    
              }
            else if (preferences.get().isNewsDownloadAllowed())
              {
                downloadNewsFeedAndNotifyAvailability(availabilityNotifier);
              }
            else
              {
                availabilityNotifier.notifyFeedCouldBeDownloaded(new DownloadConfirmation() 
                  {
                    public void confirmDownload() 
                      {
                        downloadNewsFeedAndNotifyAvailability(availabilityNotifier);
                      }
                  });
              }
          }
      }
    //@bluebook-end getNewsFeed

    /*******************************************************************************************************************
     *
     * 
     *
     ******************************************************************************************************************/
    private void ensureCacheIsInitialized (final @Nonnull AvailabilityNotifier availabilityNotifier)
      {
        if (cache.getStatus() == NEEDS_DOWNLOAD) // not initialized
          {
            try
              {
                cache.computeStatus();
              }
            catch (Exception e2) // the cache is broken
              {
                log.warn("ensureCacheIsInitialized()", e2);
                availabilityNotifier.notifyFeedUnavailable();
                return;
              }
          }
      }
    
    /*******************************************************************************************************************
     *
     * 
     *
     ******************************************************************************************************************/
    /* package */ void readNewsFeedAndNotifyAvailability (final @Nonnull AvailabilityNotifier availabilityNotifier) 
      {
        executorService.submit(new Runnable()
          {
            @Override
            public void run()
              {
                log.debug("readNewsFeedAndNotifyAvailability({})", availabilityNotifier);

                try
                  {
                    if (cache.getStatus() == DOWNLOADING)
                      {
                        waitForDownloaded();
                      }

                    doReadNewsFeedAndNotifyAvailability(availabilityNotifier); 
                  }
                catch (Exception e1)
                  {
                    log.warn("readNewsFeedAndNotifyAvailability()", e1);
                    
                    if (cache.getStatus() == NEEDS_REFRESH)
                      {
                        try 
                          {
                            availabilityNotifier.notifyCachedFeedAvailable(cache.getRssFeed());
                          }
                        catch (NotFoundException e2)
                          {
                            availabilityNotifier.notifyFeedUnavailable(); // should never happen
                          }
                      }
                    else
                      {
                        availabilityNotifier.notifyFeedUnavailable();
                      }
                  }
              }
          });
      }
    
    /*******************************************************************************************************************
     *
     * 
     *
     ******************************************************************************************************************/
    /* package */ void downloadNewsFeedAndNotifyAvailability (final @Nonnull AvailabilityNotifier availabilityNotifier) 
      {
        executorService.submit(new Runnable()
          {
            @Override
            public void run()
              {
                log.debug("downloadNewsFeedAndNotifyAvailability({})", availabilityNotifier);

                try
                  {
                    cache.download();
                    doReadNewsFeedAndNotifyAvailability(availabilityNotifier); 
                  }
                catch (Exception e)
                  {
                    log.warn("downloadNewsFeedAndNotifyAvailability()", e);
                    
                    if (cache.getStatus() == NEEDS_REFRESH)
                      {
                        doReadNewsFeedAndNotifyAvailability(availabilityNotifier);
                      }
                    else
                      {
                        availabilityNotifier.notifyFeedUnavailable();
                      }
                  }
              }
          });
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    /* package */ void doReadNewsFeedAndNotifyAvailability (final @Nonnull AvailabilityNotifier availabilityCallback)
      {
        try
          {
            final RssFeed rssFeed = readFeedFromCache();
            
            if (cache.getStatus() == NEEDS_REFRESH)
              {
                availabilityCallback.notifyCachedFeedAvailable(rssFeed);                    
              }
            else
              {
                availabilityCallback.notifyFeedAvailable(rssFeed);
              }
          }
        catch (Exception e)
          {
            log.warn("Error calling notifyFeedAvailable() or notifyCachedFeedAvailable()", e.toString());
            availabilityCallback.notifyFeedUnavailable();
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    /* package */ RssFeed readFeedFromCache()
      throws IOException
      {
        return executeInjectingRoles(new Task<RssFeed, IOException>() 
          {
            @Override @Nonnull
            public RssFeed run() 
              throws IOException 
              {
                return cache.read();
              }
          });
      }
    
    /* package */ <T, E extends Throwable> T executeInjectingRoles (final @Nonnull Task<T, E> task) 
      throws E
      {
        return ThreadLookupBinder.with(Lookups.fixed(c1, c2)).run(task);
      }

    /*******************************************************************************************************************
     *
     * Called in background to verify whether the cache needs updating. If there are new messages, posts a 
     * notification.
     *
     ******************************************************************************************************************/
    /* package */ void updateCache()
      throws Exception
      {
        log.info("updateCache()");
        cache.computeStatus();

        if (cache.getStatus() == NEEDS_DOWNLOAD)
          {
            cache.download(60000); // running in background, longer timeout
            readFeedFromCache();
            eventuallyPostUnreadNewsNotification();
          }
      }

    /*******************************************************************************************************************
     *
     * Checks whether there are unread news and eventually posts a related notification.
     *
     ******************************************************************************************************************/
    /* package */ void eventuallyPostUnreadNewsNotification()
      {
        try 
          {
            log.info("eventuallyPostUnreadNewsNotification()");

            if (!readMessagesLog.isAllRead(cache.getRssFeed()))
              {
                postUnreadNewsNotification();
              }
          }
        catch (NotFoundException e) 
          {
            log.warn("No rssFeed");
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void postUnreadNewsNotification()
      {
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected URL createUrl (final String string)
      throws IOException
      {
        return new URL(string);
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void waitForDownloaded() 
      throws IOException
      {
        synchronized (cache)
          {
            while (cache.getStatus() == DOWNLOADING)
              {
                log.debug(">>>> status is {}, waiting...", cache.getStatus());
                try
                  {
                   cache.wait();
                  } 
                catch (InterruptedException e) 
                  {
                    // FIXME
                  }
              }
          }
        
        log.debug(">>>> status is {}", cache.getStatus());
        
        if (cache.getStatus() != DOWNLOADED)
          {
            throw new IOException("Download did not complete");  
          }
      }
  }
