/***********************************************************************************************************************
 *
 * blueBill Mobile - Android - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it/mobile
 * SCM: https://java.net/hg/bluebill-mobile~android-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.mobile.news.spi;

import javax.annotation.Nonnull;
import javax.inject.Provider;
import java.util.Arrays;
import java.util.Collection;
import java.awt.event.ActionEvent;
import javax.swing.Action;
import javax.swing.AbstractAction;
import org.openide.util.Lookup;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ProxyLookup;
import it.tidalwave.util.Task;
import it.tidalwave.util.thread.annotation.ThreadConfined;
import it.tidalwave.util.ui.UserNotificationWithFeedback.Feedback;
import it.tidalwave.util.ui.FlowController;
import it.tidalwave.role.ui.PresentationModel;
import it.tidalwave.role.ui.spi.DefaultPresentationModel;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.netbeans.capabilitiesprovider.CapabilitiesProvider;
import it.tidalwave.netbeans.capabilitiesprovider.CapabilitiesProviderSupport;
import it.tidalwave.netbeans.capabilitiesprovider.ThreadLookupBinder;
import it.tidalwave.semantic.rss.RssFeed;
import it.tidalwave.bluebill.mobile.role.util.RoleRegister;
import it.tidalwave.bluebill.mobile.news.NewsService;
import it.tidalwave.bluebill.mobile.news.ui.NewsView;
import it.tidalwave.bluebill.mobile.news.ui.NewsViewController;
import it.tidalwave.bluebill.mobile.news.impl.MessagePresentationModelReadable;
import it.tidalwave.bluebill.mobile.news.impl.NewsItemActionProvider;
import lombok.extern.slf4j.Slf4j;
import static org.openide.util.NbBundle.*;
import static it.tidalwave.util.thread.ThreadType.*;
import static it.tidalwave.util.ui.UserNotification.notification;
import static it.tidalwave.util.ui.UserNotificationWithFeedback.notificationWithFeedback;
import static it.tidalwave.semantic.rss.RssFeed.RssFeed;
import static it.tidalwave.bluebill.mobile.news.Readable.Readable;

/***********************************************************************************************************************
 *
 * @stereotype Controller
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@Slf4j
public class DefaultNewsViewController implements NewsViewController
  {
    private static final Class<DefaultNewsViewController> _ = DefaultNewsViewController.class;
    
    @Nonnull
    private final NewsView view;
    
    @Nonnull
    private final FlowController flowController;
    
    private final Provider<NewsService> newsService = Locator.createProviderFor(NewsService.class);
    
    protected PresentationModel newsFeedPM;
    
    protected final RoleRegister roleRegister = new RoleRegister();
    
    /*******************************************************************************************************************
     *
     * An {@code Action} that marks all messages as read.
     *
     ******************************************************************************************************************/
    @SuppressWarnings("serial")
    /* package */ final Action markAllMessagesAsReadAction = new AbstractAction(getMessage(_, "markAllMessagesAsRead")) 
      {
        public void actionPerformed (final @Nonnull ActionEvent event) 
          {
            log.info("markAllMessagesAsReadAction.actionPerformed()");
            newsFeedPM.as(RssFeed).as(Readable).markAsRead();
            view.populate(newsFeedPM); // FIXME: fire an event instead to rerender the single message
            view.notifyAllMessagesMarkedAsRead(notification().withText(_, "allMarkedAsRead"));
            setEnabled(false);
          }
      };
    
    /*******************************************************************************************************************
     *
     * 
     *
     ******************************************************************************************************************/
    private final CapabilitiesProvider<?> messagePMRoleProvider = new CapabilitiesProviderSupport<PresentationModel>()
      {
        @Override @Nonnull
        public Collection<? extends Object> createCapabilities (final @Nonnull PresentationModel messagePM)
          {
            return Arrays.asList
              (
                new NewsItemActionProvider(messagePM),
                new MessagePresentationModelReadable(messagePM, newsFeedPM, view, markAllMessagesAsReadAction, flowController)
              );
          }
      };
    
    /*******************************************************************************************************************
     *
     * 
     *
     ******************************************************************************************************************/
    //@bluebook-begin confirmation-interaction
    private final NewsService.AvailabilityNotifier availabilityNotifier = new NewsService.AvailabilityNotifier() 
      {
        public void notifyFeedAvailable (final @Nonnull RssFeed newsFeed) 
          {
            log.debug("notifyFeedAvailable()");
            populateAndUnlockView(newsFeed);
          }

        public void notifyCachedFeedAvailable (final @Nonnull RssFeed newsFeed) 
          {
            log.debug("notifyCachedFeedAvailable()");
            populateAndUnlockView(newsFeed);
            view.notifyFeedIsCached(notification().withText(_, "obsoleteNews"));
          }

        public void notifyFeedUnavailable() 
          {
            log.debug("notifyFeedUnavailable()");
            markAllMessagesAsReadAction.setEnabled(false);
            view.unlock();
            view.notifyFeedUnavailable(notificationWithFeedback().withCaption(_, "unavailableNewsTitle")
                                                                 .withText(_, "unavailableNewsMessage")
                                                                 .withFeedback(new Feedback()
              {
                @Override
                public void onConfirm() 
                  {
                    flowController.finish();
                  }
              }));
          }

        public void notifyFeedCouldBeDownloaded (final @Nonnull NewsService.DownloadConfirmation confirmation) 
          {
            log.debug("notifyFeedCouldBeDownloaded()");
            view.confirmToDownloadNews(notificationWithFeedback().withCaption(_, "confirmDownloadTitle")
                                                                 .withText(_, "confirmDownloadMessage")
                                                                 .withFeedback(new Feedback()
              {
                @Override
                public void onConfirm()
                  {
                    confirmation.confirmDownload();
                  }

                @Override
                public void onCancel()
                  {
                    view.unlock();
                    flowController.finish();
                  }
              }));
          }
      };

    //@bluebook-end confirmation-interaction
    /*******************************************************************************************************************
     *
     * 
     *
     ******************************************************************************************************************/
    public DefaultNewsViewController (final @Nonnull NewsView view, final @Nonnull FlowController flowController)
      {
        this.view = view;
        this.flowController = flowController;
        view.bindActions(markAllMessagesAsReadAction);
      }    
    
    //@bluebook-begin confirmation-interaction
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Nonnull
    public void showNewsFeed()
      {
        log.info("showNewsFeed()");
        view.lock(notification().withText(_, "preparingNews")); 
        newsService.get().getNewsFeed(availabilityNotifier);
      }
    //@bluebook-end confirmation-interaction
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    public void cancel() 
      {
        log.info("cancel()");
        view.unlock();
        flowController.finish();
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @ThreadConfined(type=NOT_UI)
    private void populateAndUnlockView (final @Nonnull RssFeed newsFeed)
      {
        final Lookup lookup = new ProxyLookup(Lookups.fixed(messagePMRoleProvider), roleRegister.getLookup());
        ThreadLookupBinder.with(lookup).run(new Task<Void, RuntimeException>() 
          {
            @Override
            public Void run() 
              {
                log.debug("populateAndUnlockView({})", newsFeed);
                markAllMessagesAsReadAction.setEnabled(!newsFeed.as(Readable).isRead());
                newsFeedPM = new DefaultPresentationModel(newsFeed);
                view.populate(newsFeedPM);
                view.unlock();
                return null;
              }
          });
      }
  }
